import json
import re
import requests
from hubu.base import Base


class HumbleBundle(Base):
    def __init__(self, auth_session_id):
        super(__class__, self).__init__()
        self.__cookies = dict(_simpleauth_sess=auth_session_id)

    def purchases(self):
        response = requests.get('https://www.humblebundle.com/home/purchases', cookies=self.__cookies)
        keys = json.loads(re.search(r'"gamekeys"\s*:\s*(.*)', response.text).groups()[0][:-1])
        return [Purchase(key, self.__cookies) for key in keys]

class Purchase(Base):
    def __init__(self, key, cookies):
        super(__class__, self).__init__()
        self.key = key
        self.cookies = cookies
        self.info = None
        self.sanitized_name = None

    def _download_info(self):
        url = f'https://www.humblebundle.com/api/v1/order/{self.key}?all_tpkds=true'
        response = requests.get(url, cookies=self.cookies)
        self.info = json.loads(response.text)
        self.sanitized_name = self.slugify(self.info['product']['human_name'])

    def products(self):
        if self.info is None:
            self._download_info()
        for product in self.info['subproducts']:
            product['slugify_name'] = self.slugify(product['human_name'])
            yield product



import boto3
import logging
import os
from botocore.client import ClientError


class BucketClient:
    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)

        options = {}

        if 'ENDPOINT_URL' in os.environ:
            options['endpoint_url'] = os.environ.get('ENDPOINT_URL')

        if 'AWS_ACCESS_KEY_ID' in os.environ:
            options['aws_access_key_id'] = os.environ.get('AWS_ACCESS_KEY_ID')

        if 'AWS_SECRET_ACCESS_KEY' in os.environ:
            options['aws_secret_access_key'] = os.environ.get('AWS_SECRET_ACCESS_KEY')

        if 'BUCKET_NAME' in os.environ:
            self.bucket_name = os.environ.get('BUCKET_NAME')
        else:
            raise Exception('BUCKET_NAME is not defined!')

        self.__s3 = boto3.client('s3', **options)

        if not self.__bucket_exists():
            raise Exception("Bucket check failed")

    def __bucket_exists(self):
        try:
            self.__s3.head_bucket(Bucket=self.bucket_name)
            self.log.info('Bucket exists')
            return True
        except ClientError:
            self.log.error('Bucket does not exist')
            return False
        except Exception as e:
            self.log.error('Something went wrong', e)
            return False

    def delete_object(self, path):
        self.log.info(f'Delete file: {path}')
        self.__s3.delete_object(Bucket=self.bucket_name, Key=path)

    def put_object(self, body, path):
        self.__s3.put_object(Bucket=self.bucket_name,
                             Key=path,
                             Body=body)

    def upload_fileobj(self, body, path, callback=None):
        self.__s3.upload_fileobj(Bucket=self.bucket_name,
                                 Key=path,
                                 Fileobj=body,
                                 Callback=callback)

    def file_size(self, path):
        """Return key's size, or zero if does not exist"""
        try:
            response = self.__s3.head_object(Bucket=self.bucket_name,
                                             Key=path)
            return response['ContentLength']
        except:
            return 0



import urllib
import tempfile
import threading
import time
from hubu.base import ThreadedBase
from hubu.bucket_client import BucketClient

class Downloader(ThreadedBase):
    def __init__(self, thread_id, manager):
        super(__class__, self).__init__(thread_id)
        self.thread_id = thread_id
        self.manager = manager
        self._target_size = 0
        self._transferred_bytes = 0
        self.action = 'nothing'
        self.__bucket_client = None

    def s3(self):
        if self.__bucket_client is None:
            self.log.info('Create bucket client')
            self.__bucket_client = BucketClient()
        return self.__bucket_client

    def progress(self):
        if self._transferred_bytes == 0:
            return 0.0
        return self._transferred_bytes / self._target_size

    def action_icon(self):
        actions = {
            'download': '↓',
            'upload': '↑'
        }
        return actions.get(self.action, ' ')


    def run(self):
        self.log.info(f'--- Start #{self.thread_id}')
        self.process()
        self.log.info(f'--- Stop #{self.thread_id}')

    def process(self):
        while True:
            self.action = 'nothing'
            self.manager.lock().acquire()
            if self.manager.queue().empty():
                self.manager.lock().release()
                time.sleep(1)
                continue

            item = self.manager.queue().get()
            self.manager.lock().release()
            if item == -1:
                break

            size = self.s3().file_size(item.path())
            if size == item.size:
                self.log.info(f'{item.path()} already in there')
                self.manager.queue().task_done()
                continue

            if size != 0 and size != item.size:
                self.log.info(f'{item.path()} already in there, but with wrong size')
                self.s3().delete_object(item.path())

            self._target_size = item.size
            self._transferred_bytes = 0
            fp = tempfile.TemporaryFile()
            self.save(source=item.url, dest=fp)
            fp.seek(0)
            self.upload(source=fp, dest=item.path())
            fp.close()

            self.manager.queue().task_done()

    def save(self, source, dest, blockSize=200):
        self.action = 'download'
        try:
            url = urllib.request.urlopen(source)
        except:
            print("SKIP", source, fullPath)
            return

        block_size = (1024*blockSize)
        while True:
            buf = url.read(block_size)
            if not buf:
                break
            self._transferred_bytes += len(buf)
            dest.write(buf)

    def upload(self, source, dest):
        self.action = 'upload'
        self.log.info(f"Upload {dest}")
        self._transferred_bytes = 0
        def upload_progress(tx):
            self._transferred_bytes = self._transferred_bytes + tx
        self.s3().upload_fileobj(source, dest, callback=upload_progress)



import logging
import re
import threading
import unicodedata

class Base(object):
    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)

    def slugify(self, value):
        """
        Convert to ASCII. Convert spaces to hyphens.
        Remove characters that aren't alphanumerics, underscores, or hyphens.
        Convert to lowercase. Also strip leading and trailing whitespace.

        Source: https://docs.djangoproject.com/en/2.1/_modules/django/utils/text/#slugify
        """
        value = str(value)
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value).strip().lower()
        return re.sub(r'[-\s]+', '-', value)


class ThreadedBase(threading.Thread):
    def __init__(self, thread_id):
        threading.Thread.__init__(self)
        self.log = logging.getLogger(f'{self.__class__.__name__}#{thread_id}')



import boto3
import logging
import os
import time
from urllib.parse import urlparse
from hubu.download_manager import DownloadManager
from hubu.downloadable import Downloadable
from hubu.humble_bundle import HumbleBundle


if 'HUMBLE_BUNDLE_SESSION' not in os.environ:
    raise Exception('HUMBLE_BUNDLE_SESSION is not defined')
sess = os.environ.get('HUMBLE_BUNDLE_SESSION')


def main():
    logging.basicConfig(level=getattr(logging, 'INFO', None),
                        format='%(asctime)s | %(levelname)-8s | %(name)-20s | %(message)s')

    with DownloadManager(workers=5) as mng:
        acceptable_platforms = [
                'ebook', 'video'
        ]
        sum_size = 0
        purchases = HumbleBundle(sess).purchases()
        for purchase in purchases:
            for product in purchase.products():
                for d in product['downloads']:
                    if d['platform'] not in acceptable_platforms:
                        continue
                    for ds in d['download_struct']:
                        if 'name' not in ds or 'url' not in ds:
                            continue

                        if ds['name'] == 'PDF (HD)':
                            ds['name'] = 'PDF'
                        if ds['name'] == 'PDF (HQ)':
                            ds['name'] = 'PDF'

                        url = urlparse(ds['url']['web'])
                        filename = os.path.basename(url.path)
                        sum_size += ds['file_size']
                        item = Downloadable(bundle=purchase.sanitized_name,
                                            name=product['slugify_name'],
                                            platform=d['platform'],
                                            filename=filename,
                                            url=ds['url']['web'],
                                            size=ds['file_size'])

                        mng.add(item)

        while not mng.queue().empty():
            mng.status_line()
            time.sleep(1)
            pass



"""
# Fill
log.info('prefill')
queueLock.acquire()
main()
queueLock.release()

while not downloadQueue.empty():
    line = " ".join(["  {:6.2f}%  ".format(t.progress() * 100) for t in threads])
    print(f"\r[{downloadQueue.qsize()}] {line}", end="")
    #time.sleep(1)
    pass

"""




import os
from hubu.base import Base


class Downloadable(Base):
    def __init__(self, bundle, name, platform, filename, url, size):
        super(__class__, self).__init__()
        self.download_path = f"{bundle}/{platform}/{name}/{filename}"
        self.url = url
        self.size = size

    def path(self, base=''):
        return os.path.join(base, self.download_path)

    def __lt__(self, other):
        return self.size < getattr(other, 'size', other)
    def __gt__(self, other):
        return self.size > getattr(other, 'size', other)
    def __aq__(self, other):
        return self.size == getattr(other, 'size', other)




import queue
import threading
from hubu.base import Base
from hubu.downloader import Downloader

class DownloadManager(Base):
    def __init__(self, workers):
        super(__class__, self).__init__()
        self.number_of_workers = workers
        self.__queue = queue.PriorityQueue()
        self.__lock = threading.Lock()
        self.threads = []

    def __enter__(self):
        self.__start_workers()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__clear_workers()

    def __start_workers(self):
        self.log.info('start workers')
        for i in range(0, self.number_of_workers):
            t = Downloader(i, self)
            t.start()
            self.threads.append(t)

    def __clear_workers(self):
        self.log.info('Clear workers')
        for i in range(0, self.number_of_workers):
            self.__queue.put(-1)
        for t in self.threads:
            t.join()

    def add(self, item):
        self.lock().acquire()
        self.queue().put(item)
        self.lock().release()

    def queue(self):
        return self.__queue

    def lock(self):
        return self.__lock

    def status_line(self):
        line = " ".join(["  {:s}{:6.2f}%  ".format(t.action_icon(), t.progress() * 100) for t in self.threads])
        self.log.info(f"[{self.queue().qsize()}] {line}")

